import asyncio
import json
import secrets
from typing import List, Tuple, Optional
from uuid import uuid4
import typer
from enum import Enum
from opal_common.schemas.security import AccessTokenRequest, PeerType
from opal_common.schemas.data import DataSourceEntry, DataUpdate
from datetime import timedelta


class SecretFormat(str, Enum):
    hex = "hex"
    bytes = "bytes"
    urlsafe = "urlsafe"


def generate_secret(size: int = typer.Option(32, help="size in bytes of the secret"),
                    format: SecretFormat = SecretFormat.urlsafe):
    if format == SecretFormat.hex:
        res = secrets.token_hex(size)
    elif format == SecretFormat.bytes:
        res = repr(secrets.token_bytes(size))
    else:
        res = secrets.token_urlsafe(size)

    typer.echo(res)


def obtain_token(
        master_token: str = typer.Argument(...,help="The master token secret the OPAL-server was initialized with", envvar="OPAL_MASTER_TOKEN"),
        server_url: str = typer.Option("http://localhost:7002", help="url of the OPAL-server to obtain the token from"),
        type: PeerType = PeerType("client"),
        ttl: Tuple[int, str] = typer.Option(
            (365, "days"), help="Time-To-Live / experation for the token in `<int> <str>` e.g. `365 days`, or `1000000 milliseconds` "),
        claims: str = typer.Option(
            "{}", help="claims to to include in the returned signed JWT as a JSON string", callback=lambda x: json.loads(x)),
        just_the_token: bool = typer.Option(
            True, help="Should the command return only the cryptographic token, or the full JSON object"),
):
    """
    Obtain a secret JWT (JSON-Web-Token) from the server, to be used by clients or data sources for authentication
    Using the master token (as assigned to the server as OPAL_AUTH_MASTER_TOKEN)
    """

    from aiohttp import ClientSession

    server_url = f"{server_url}/token"
    ttl_number, ttl_unit = ttl
    ttl = timedelta(**{ttl_unit: ttl_number})

    async def fetch():
        async with ClientSession(headers={"Authorization": f"bearer {master_token}"}) as session:
            details = AccessTokenRequest(type=type, ttl=ttl, claims=claims).json()
            res = await session.post(server_url, data=details, headers={'content-type': 'application/json'})
            data = await res.json()
            if just_the_token:
                return data["token"]
            else:
                return data

    res = asyncio.run(fetch())
    typer.echo(res)


def publish_data_update(
        token: Optional[str] = typer.Argument(None,help="the JWT obtained from the server for authentication (see obtain-token command)",
                                    envvar="OPAL_CLIENT_TOKEN"),
        server_url: str = typer.Option("http://localhost:7002",
                                        help="url of the OPAL-server to send the update through"),
        server_route: str = typer.Option("/data/config",
                                        help="route in the server for update"),
        reason: str = typer.Option("",
                                   help="The reason for the update"),
        entries: str = typer.Option("[]", "--entries", "-e",
                                    help="Pass in the the DataUpdate entries as JSON",
                                    callback=lambda x: json.loads(x)),

        src_url: str = typer.Option(None,
                                    help="[SINGLE-ENTRY-UPDATE] url of the data-source this update relates to, which the clients should approach"),
        topics: List[str] = typer.Option(None, "--topic","-t",
                                    help="[SINGLE-ENTRY-UPDATE] [List] topic (can several) for the published update (to be matched to client subscriptions)"),
        src_config: str = typer.Option("{}",
                                    help="[SINGLE-ENTRY-UPDATE] Fetching Config as JSON",  callback=lambda x: json.loads(x)),
        dst_path: str = typer.Option("",
                                    help="[SINGLE-ENTRY-UPDATE] Path the client should set this value in its data-store"),
        save_method: str = typer.Option("PUT",
                                    help="[SINGLE-ENTRY-UPDATE] How the data should be saved into the give dst-path")):
    """
    Publish a DataUpdate through an OPAL-server (indicated by --server_url).

    [SINGLE-ENTRY-UPDATE]
        Send a single update DataSourceEntry via the --src-url, --src-config, --topics, --dst-path, --save-method
        must include --src-url to use this flow.
    [Multiple entries]
        Set DataSourceEntires as JSON (via --entries)
        if you include a single entry as well- it will be merged into the given JSON
    """
    from aiohttp import ClientSession, ClientResponse

    if not isinstance(entries, list):
        typer.secho("Bad input for --entires was ignored", fg="red")
        entries = []

    entries: List[DataSourceEntry]

    # single entry update
    if src_url is not None:
        entry = DataSourceEntry(url=src_url, topics=topics, dst_path=dst_path, save_method=save_method, config=src_config)
        entries.append(entry)

    server_url = f"{server_url}{server_route}"
    update = DataUpdate(entries=entries, reason=reason)

    async def publish_update():
        headers = {'content-type': 'application/json'}
        if token is not None:
            headers.update({"Authorization": f"bearer {token}"})
        async with ClientSession(headers=headers) as session:
            body = update.json()
            res = await session.post(server_url, data=body)
            return res

    async def get_response_text(res: ClientResponse):
        return await res.text()

    typer.echo(f"Publishing event:")
    typer.secho(f"{str(update)}", fg="cyan")
    res = asyncio.run(publish_update())

    if res.status == 200:
        typer.secho("Event Published Successfully", fg="green")
    else:
        typer.secho("Event publishing failed with status-code - {res.status}", fg="red")
        text = asyncio.run(get_response_text(res))
        typer.echo(text)



all_commands = [obtain_token, generate_secret, publish_data_update]




import inspect
from typing import Any, Callable, List, Type
from decouple import  text_type, undefined


class FromStr:
    """
    Placeholder for values parsed from strings into more complex objects
    """
    def __init__(self, type, cast):
        self._type = type
        self.cast = cast

    def __call__(self, arg) -> Any:
        if arg is not undefined:
            return self.cast(arg)
        else:
            return undefined

    @property
    def __name__(self) -> str:
        if hasattr(self._type,'__name__'):
            return f"<{self._type.__name__}>"
        else: 
            return repr(self._type)

def no_cast(value):
    return value
class ConfiEntry:
    key:str
    type: Type
    default: Any
    description:str
    cast:Callable
    kwargs: dict 
    flags:List[str]
    value: Any

    def __init__(self, key, default=undefined, description=None, cast=no_cast, type=str,index=-1, flags:List[str]=None, **kwargs) -> None:
        self.key = key
        # sorting index
        self.index = index
        self.default = default
        self.description = description
        self.cast = cast
        self.type = type
        self.kwargs = kwargs
        self.flags = flags
        self.value = undefined

    def get_cli_type(self):
        if self.type in {str, int, float, list, dict, bool}:
            return self.type
        else: 
            return FromStr(self.type, self.cast)

    def get_cli_option_kwargs(self):
        res = {
            "type": self.get_cli_type(),
        }
        if self.description is not None:
            res['help'] = self.description
        if self.default is not undefined:
            res['default'] = self.value if self.value is not undefined else self.default
            res['show_default'] = self.default
        return res
        
        
class ConfiDelay:
    """
    Delay loaded confi entry default values
    """

    def __init__(self, value, index=-1) -> None:
        self._value = value
        # sorting index
        self.index = index

    @property
    def value(self): 
        return self._value

    def eval(self, config=None):
        values = { k:v.value for k,v in config.entries.items()} if config else {}
        if isinstance(self._value, str):
            return self._value.format(**values)
        if callable(self._value):
            callargs = inspect.getcallargs(self._value)
            args = {k: values.get(k, undefined) for k,v in callargs.items()}
            return self._value(**args)

    def __repr__(self) -> str:
        try:
            return f"<Delayed {self.eval()}>"
        except:
            return f"<Delayed {self._value}>" 



from datetime import datetime, timedelta
from typing import Optional
from uuid import UUID

import jwt
from opal_common.authentication.verifier import JWTVerifier
from jwt.api_jwk import PyJWK

from opal_common.authentication.types import JWTAlgorithm, JWTClaims, PrivateKey, PublicKey
from opal_common.logger import logger

class InvalidJWTCryptoKeysException(Exception):
    """
    raised when JWT signer provided with invalid crypto keys
    """
    pass

class JWTSigner(JWTVerifier):
    """
    given cryptographic keys, signs and verifies jwt tokens.
    """
    def __init__(
        self,
        private_key: Optional[PrivateKey],
        public_key: Optional[PublicKey],
        algorithm: JWTAlgorithm,
        audience: str,
        issuer: str,
    ):
        """
        inits the signer if and only if the keys provided to __init__
        were generate together are are valid. otherwise will throw.

        JWT signer can be initialized with empty keys (None),
        in which case signer.enabled == False.

        This allows opal to run both in secure mode (which keys, requires jwt authentication)
        and in insecure mode (good for development and running locally).

        Args:
            private_key (PrivateKey): a valid private key or None
            public_key (PublicKey): a valid public key or None
            algorithm (JWTAlgorithm): the jwt algorithm to use
                (possible values: https://pyjwt.readthedocs.io/en/stable/algorithms.html)
            audience (string): the value for the aud claim: https://tools.ietf.org/html/rfc7519#section-4.1.3
            issuer (string): the value for the iss claim: https://tools.ietf.org/html/rfc7519#section-4.1.1
        """
        super().__init__(
            public_key=public_key,
            algorithm=algorithm,
            audience=audience,
            issuer=issuer
        )
        self._private_key = private_key
        self._verify_crypto_keys()

    def _verify_crypto_keys(self):
        """
        verifies whether or not valid crypto keys were provided to the signer.
        if both keys are valid, encodes and decodes a JWT to make sure the keys match.

        if both private and public keys are valid and are matching => signer is enabled
        if both private and public keys are None => signer is disabled (self.enabled == False)
        if only one key is valid/not-None => throws ValueError
        any other case => throws ValueError
        """
        if self._private_key is not None and self._public_key is not None:
            # both keys provided, let's make sure these keys were generated correctly
            token = jwt.encode({"some": "payload"}, self._private_key, algorithm=self._algorithm)
            try:
                jwt.decode(token, self._public_key, algorithms=[self._algorithm])
            except jwt.PyJWTError as exc:
                logger.info("JWT Signer key verification failed with error: {err}", err=repr(exc))
                raise InvalidJWTCryptoKeysException("private key and public key do not match!") from exc
            # save jwk
            self._jwk: PyJWK = PyJWK.from_json(self.get_jwk(), algorithm=self._algorithm)
        elif (self._private_key != self._public_key) and (self._private_key is None or self._public_key is None):
            raise ValueError("JWT Signer not valid, only one of private key / public key pair was provided!")
        elif self._private_key is None and self._public_key is None:
            # valid situation, running in dev mode and api security is off
            self._disable()
        else:
            raise ValueError("Invalid JWT Signer input!")

    def sign(self, sub: UUID, token_lifetime: timedelta, custom_claims: dict = {}) -> str:
        payload = {}
        issued_at = datetime.utcnow()
        expire_at = issued_at + token_lifetime
        payload = {
            "iat": issued_at,
            "exp": expire_at,
            "aud": self._audience,
            "iss": self._issuer,
            "sub": sub.hex,
        }
        if custom_claims:
            payload.update(custom_claims)

        headers = {}
        if self._jwk.key_id is not None:
            headers = {"kid": self._jwk.key_id}
        return jwt.encode(payload, self._private_key, algorithm=self._algorithm, headers=headers)
